{{
Kalman filter manager

This code runs a Kalman filter.  It used to be on its own cog, but was fast enough to move into
the main thread.  It uses readings from a dual-axis gyro and a tri-axis accelerometer to correct
gyro drift and return an estimate of both the current orientation in X & Y, as well as unbiased
gyro readings in X & Y (pitch and roll only).

}}


CON
  One                     = 1.0
  AccelToRadians          = 7.6699039394282061485904379474597e-4    'pi / 12 bits
  RadiansTo12Bit          = 1303.7972938088065906186957895476 
  HzToSec                 = 1.0 / 80_000_000.0        

  'Gyro range is 0 to 3.0v, not 3.3v, so 3723 is +500, and 0 is -500, with center at ~1862
  '1862 @ 12 bit is 500, so 670.32 = 180  (pi / 670.32 = GyroToRadians)

  'I had to add the * 2.0 to get the gyro estimate to coincide with the accelerometer measurement,
  'but I still haven't had time to work out why.       
  GyroToRadians           = 0.0046867058324230117532859580249426 * 2.0 

VAR
  'long  Stack[128]                                      ' Probably don't need nearly this much, but I have lots
  long  xAngleInt, yAngleInt

  long  CurTime, PrevTime, DeltaTime, fDelta
  long  AccelX, AccelY, fAccelX, fAccelY  
  long  GyroX, GyroY, fGyroX, fGyroY
  long  GyroXUnb, GyroYUnb, fGyroXUnb, fGyroYUnb

  long  GyroXCenter, GyroYCenter
  long  AccelXCenter, AccelYCenter

  long xAngle, xBias
  long xP00, xP01, xP10, xP11

  long yAngle, yBias
  long yP00, yP01, yP10, yP11

  long QAngle, QGyro, RAngle  


OBJ
  IMU   : "H48CAccelMod.spin"
  f     : "Float32.spin"                        

PUB Start
  IMU.start( 20, 21, 22 )                               ' Initialize the accelerometer / gyro
  f.start  

  KalmanInit( 0.0001, 0.007, 1.5 )                       ' Initialize the 2D Kalman filter

  xAngleInt := yAngleInt := 0

  waitcnt( cnt + 4_000_000 )                            ' Wait 1/2 second before initializing gyros                        

  'Compute initial gyro center values
  GyroXCenter := 0
  GyroYCenter := 0
  AccelXCenter := 0
  AccelYCenter := 0
  repeat 16
    GyroXCenter += IMU.Gx
    GyroYCenter += IMU.Gy
    AccelXCenter += IMU.ThetaB
    AccelYCenter += IMU.ThetaC
    waitcnt( cnt + 20_000 )

  GyroXCenter := (GyroXCenter + 8) >> 4
  GyroYCenter := (GyroYCenter + 8) >> 4
  AccelXCenter := (AccelXCenter + 8) >> 4
  AccelYCenter := (AccelYCenter + 8) >> 4

  PrevTime := cnt

  'cognew( KalmanManager, @Stack )
  return


PUB GetX                                                ' Get the current X-roll value (0-4095)
  return xAngleInt
               
PUB GetY                                                ' Get the current Y-pitch value (0-4095)
  return yAngleInt

PUB GetGyroX                                            ' Get the unbiased (drift corrected) gyro X value
  return GyroXUnb

PUB GetGyroY                                            ' Get the unbiased (drift corrected) gyro Y value
  return GyroYUnb 

PUB GetAccelX                                           ' Get the raw, noisy, accelerometer X angle (for testing)
  return AccelX

PUB GetAccelY                                           ' Get the raw, noisy, accelerometer Y angle (for testing)                        
  return AccelY


PUB Update                                              

  'Read the accelerometer
  AccelX := -(IMU.ThetaB - AccelXCenter)
  AccelY := (IMU.ThetaC - AccelYCenter)
   
  'Read the gyros
  GyroX := (IMU.Gx - GyroXCenter)              
  GyroY := -(IMU.Gy - GyroYCenter)
   
  'Convert the values to float (radians)
  fAccelX := f.FMul( f.FFloat(AccelX), AccelToRadians ) 
  fAccelY := f.FMul( f.FFloat(AccelY), AccelToRadians )
   
  fGyroX := f.FMul( f.FFloat(GyroX), GyroToRadians ) 
  fGyroY := f.FMul( f.FFloat(GyroY), GyroToRadians ) 
   
   
  'Compute the time delta
  CurTime := cnt
  DeltaTime := CurTime - PrevTime
  PrevTime := CurTime

  'Convert the delta to seconds    
  fDelta := f.FFloat(DeltaTime)
  fDelta := f.FMul(fDelta, HzToSec)

  'Run the filter   
  KalmanPredict( fGyroX, fGyroY, fDelta )
  KalmanUpdate( fAccelX, fAccelY )

  'Produce the results   
  xAngleInt := f.FRound(f.FMul(xAngle, RadiansTo12Bit)) 
  yAngleInt := f.FRound(f.FMul(yAngle, RadiansTo12Bit))
   
  GyroXUnb := f.FRound(f.FMul(fGyroXUnb, RadiansTo12Bit)) 
  GyroYUnb := f.FRound(f.FMul(fGyroYUnb, RadiansTo12Bit))
   


PRI KalmanInit( QAngle_, QGyro_, RAngle_ ) 
  {{ Initialize a new Kalman Filter object }}
  {{ QAngle = Trust in accelerometer measurements }}
  {{ QGyro = Trust in gyro measurements }}
  {{ RAngle = expected noise in accelerometer measurements in radians }} 

  xAngle := xBias := GyroXUnb := GyroYUnb := 0
  xP00 := xP01 := xP10 := xP11 := 0

  yAngle := yBias := 0
  yP00 := yP01 := yP10 := yP11 := 0
  
  QAngle := QAngle_
  QGyro  := QGyro_
  RAngle := RAngle_


PRI KalmanPredict( RateX, RateY, dt ) | temp, negdt, negdtXp11, negdtYp11, QAngledt, QGyrodt

  negdt := f.FNeg(dt)     
  negdtXp11 := f.FMul(negdt, xP11) 
  negdtYp11 := f.FMul(negdt, yP11)
  QAngledt := f.FMul(QAngle, dt) 
  QGyrodt := f.FMul(QGyro, dt) 


  'xAngle += dt * (x - xBias );
  fGyroXUnb := f.FSub(RateX, xBias) 
  xAngle := f.FAdd( xAngle, f.FMul(dt, fGyroXUnb) )   

  'xP00 +=  -dt * (xP10 + xP01) + xQ_angle * dt;
  xP00 := f.FAdd( xP00, f.FAdd(f.FMul(negdt, f.FAdd(xP10, xP01)), QAngledt))

  'xP01 +=  -dt * xP_11;
  xP01 := f.FAdd(xP01, negdtXp11)

  'xP10 +=  -dt * xP_11;
  xP10 := f.FAdd(xP10, negdtXp11)

  'xP11 +=  xQ_gyro * dt;
  xP11 := f.FAdd(xP11, QGyrodt)


  'yAngle += dt * (y - yBias );
  fGyroYUnb := f.FSub(RateY, yBias)
  yAngle := f.FAdd( yAngle, f.FMul(dt, fGyroYUnb) )   

  'yP00 +=  -dt * (yP10 + yP01) + QAngle * dt;
  yP00 := f.FAdd( yP00, f.FAdd(f.FMul(negdt, f.FAdd(yP10, yP01)), QAngledt))

  'yP01 +=  -dt * yP_11;
  yP01 := f.FAdd(yP01, negdtYp11)

  'yP10 +=  -dt * yP_11;
  yP10 := f.FAdd(yP10, negdtYp11)

  'xP11 +=  xQ_gyro * dt;
  yP11 := f.FAdd(yP11, QGyrodt)




PRI KalmanUpdate( AngleX , AngleY ) | err, S, K0, K1

  err := f.FSub(AngleX, xAngle)
  S := f.FAdd(xP00, RAngle)

  K0 := f.FDiv(xP00, S)
  K1 := f.FDiv(xP10, S)
   
  xAngle := f.FAdd(xAngle, f.FMul(K0, err))
  xBias  := f.FAdd( xBias, f.FMul(K1, err))
   
  xP10 := f.FSub(xP10, f.FMul(K1, xP00))
  xP11 := f.FSub(xP11, f.FMul(K1, xP01))
  xP00 := f.FSub(xP00, f.FMul(K0, xP00))
  xP01 := f.FSub(xP01, f.FMuL(K0, xP01))


  err := f.FSub(AngleY, yAngle)
  S := f.FAdd(yP00, RAngle)

  K0 := f.FDiv(yP00, S)
  K1 := f.FDiv(yP10, S)
   
  yAngle := f.FAdd(yAngle, f.FMul(K0, err))
  yBias  := f.FAdd( yBias, f.FMul(K1, err))
   
  yP10 := f.FSub(yP10, f.FMul(K1, yP00))
  yP11 := f.FSub(yP11, f.FMul(K1, yP01))
  yP00 := f.FSub(yP00, f.FMul(K0, yP00))
  yP01 := f.FSub(yP01, f.FMul(K0, yP01))
  